import re
import string
from typing import Any, Dict, List, Optional
from unittest.mock import sentinel

import pytest
from grimp.adaptors.graph import ImportGraph

from importlinter.application.app_config import settings
from importlinter.application.ports.building import GraphBuilder
from importlinter.application.use_cases import (
    FAILURE,
    SUCCESS,
    create_report,
    lint_imports,
)
from importlinter.application.user_options import UserOptions
from tests.adapters.building import FakeGraphBuilder
from tests.adapters.printing import FakePrinter
from tests.adapters.timing import FakeTimer
from tests.adapters.user_options import (
    ExceptionRaisingUserOptionReader,
    FakeUserOptionReader,
)

SOME_CACHE_DIR = "/path/to/some/cache/dir"


class TestCheckContractsAndPrintReport:
    def test_all_successful(self):
        self._configure(
            contracts_options=[
                {"type": "always_passes", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        result = lint_imports()

        assert result == SUCCESS

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT
            Contract bar KEPT

            Contracts: 2 kept, 0 broken.
            """
        )

    def test_invalid_contract(self):
        self._configure(
            contracts_options=[
                {
                    "type": "fields",
                    "name": "Contract foo",
                    "single_field": ["one", "two"],
                    "multiple_field": "one",
                    "import_field": "foobar",
                },
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        result = lint_imports()

        assert result == FAILURE

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            Contract "Contract foo" is not configured correctly:
                single_field: Expected a single value, got multiple values.
                import_field: Must be in the form "package.importer -> package.imported".
                required_field: This is a required field.
            """
        )

    def test_one_failure(self):
        self._configure(
            contracts_options=[
                {"type": "always_fails", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        result = lint_imports()

        assert result == FAILURE

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo BROKEN
            Contract bar KEPT

            Contracts: 1 kept, 1 broken.


            ----------------
            Broken contracts
            ----------------

            Contract foo
            ------------

            This contract will always fail.
            """
        )

    def test_limit_contracts(self):
        self._configure(
            contracts_options=[
                {"type": "always_passes", "id": "blue", "name": "Contract blue"},
                {"type": "always_passes", "id": "green", "name": "Contract green"},
                {"type": "always_fails", "id": "yellow", "name": "Contract yellow"},
                {"type": "always_passes", "id": "purple", "name": "Contract purple"},
            ]
        )

        result = lint_imports(limit_to_contracts=("green", "purple"))

        assert result == SUCCESS
        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract green KEPT
            Contract purple KEPT

            Contracts: 2 kept, 0 broken.
            """
        )

    def test_warnings(self):
        self._configure(
            contracts_options=[
                {
                    "type": "always_passes",
                    "name": "Contract foo",
                    "warnings": ["Some warning.", "Another warning."],
                },
                {
                    "type": "always_fails",
                    "name": "Contract bar",
                    "warnings": ["A third warning."],
                },
            ]
        )

        result = lint_imports()

        assert result == FAILURE

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT (2 warnings)
            Contract bar BROKEN (1 warning)

            Contracts: 1 kept, 1 broken.

            --------
            Warnings
            --------

            Contract foo
            ------------

            - Some warning.
            - Another warning.

            Contract bar
            ------------

            - A third warning.


            ----------------
            Broken contracts
            ----------------

            Contract bar
            ------------

            This contract will always fail.
            """
        )

    def test_timings(self):
        timer = FakeTimer()
        timer.setup(tick_duration=5, increment=10)
        self._configure(
            contracts_options=[
                {
                    "type": "always_passes",
                    "name": "Contract foo",
                },
                {
                    "type": "always_passes",
                    "name": "Contract bar",
                    "warnings": ["A warning."],
                },
            ],
            timer=timer,
        )

        lint_imports(show_timings=True)

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            Building graph took 5.0s.

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT [15s]
            Contract bar KEPT (1 warning) [25s]

            Contracts: 2 kept, 0 broken.

            --------
            Warnings
            --------

            Contract bar
            ------------

            - A warning.
            """
        )

    @pytest.mark.parametrize(
        "cache_dir, expected_graph_building_output",
        (
            (
                "/path/to/cache",
                "Building import graph (cache directory is /path/to/cache)...",
            ),
            (None, "Building import graph (with caching disabled)..."),
            (
                sentinel.not_supplied,
                f"Building import graph (cache directory is {SOME_CACHE_DIR})...",
            ),
        ),
    )
    @pytest.mark.parametrize(
        "verbose, expected_output_template",
        [
            (
                True,
                """
                =============
                Import Linter
                =============

                Verbose mode.
                {{ graph building output }}
                Built graph in 5.0s.
                Checking Contract foo...
                Hello from the noisy contract!
                Contract foo KEPT [15s]
                Checking Contract bar...
                Contract bar KEPT [25s]

                ---------
                Contracts
                ---------

                Analyzed 26 files, 10 dependencies.
                -----------------------------------

                Contract foo KEPT
                Contract bar KEPT

                Contracts: 2 kept, 0 broken.
                """,
            ),
            (
                False,
                """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT
            Contract bar KEPT

            Contracts: 2 kept, 0 broken.
            """,
            ),
        ],
    )
    def test_verbose_mode(
        self,
        verbose,
        expected_output_template,
        cache_dir,
        expected_graph_building_output,
    ):
        timer = FakeTimer()
        timer.setup(tick_duration=5, increment=10)
        self._configure(
            contracts_options=[
                {"type": "noisy", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ],
            timer=timer,
        )

        kwargs = dict(verbose=verbose, is_debug_mode=True)
        if cache_dir is not sentinel.not_supplied:
            kwargs["cache_dir"] = cache_dir

        lint_imports(**kwargs)

        expected_output = expected_output_template.replace(
            "{{ graph building output }}", expected_graph_building_output
        )
        settings.PRINTER.pop_and_assert(expected_output)

    @pytest.mark.parametrize(
        "passed_arg, expected_cache_dir",
        (
            (sentinel.not_supplied, SOME_CACHE_DIR),
            ("/path/to/cache", "/path/to/cache"),
            (None, None),
        ),
    )
    def test_cache_dir_is_passed_correctly(self, passed_arg, expected_cache_dir):
        builder = FakeGraphBuilder()
        self._configure(contracts_options=[], graph_builder=builder)

        if passed_arg is sentinel.not_supplied:
            lint_imports()
        else:
            lint_imports(cache_dir=passed_arg)

        assert builder.build_arguments["cache_dir"] == expected_cache_dir

    def test_forbidden_import(self):
        """
        Tests the ForbiddenImportContract - a simple contract that
        looks at the graph.
        """
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.foo",
            imported="mypackage.bar",
            line_number=8,
            line_contents="from mypackage import bar",
        )
        graph.add_import(
            importer="mypackage.foo",
            imported="mypackage.bar",
            line_number=16,
            line_contents="from mypackage.bar import something",
        )
        self._configure(
            contracts_options=[
                {"type": "always_passes", "name": "Contract foo"},
                {
                    "type": "forbidden",
                    "name": "Forbidden contract one",
                    "importer": "mypackage.foo",
                    "imported": "mypackage.bar",
                },
                {
                    "type": "forbidden",
                    "name": "Forbidden contract two",
                    "importer": "mypackage.foo",
                    "imported": "mypackage.baz",
                },
            ],
            graph=graph,
        )

        result = lint_imports()

        assert result == FAILURE

        # Expecting 28 files (default graph has 26 modules, we add 2).
        # Expecting 11 dependencies (default graph has 10 imports, we add 2,
        # but it counts as 1 as it's between the same modules).
        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            ---------
            Contracts
            ---------

            Analyzed 28 files, 11 dependencies.
            -----------------------------------

            Contract foo KEPT
            Forbidden contract one BROKEN
            Forbidden contract two KEPT

            Contracts: 2 kept, 1 broken.


            ----------------
            Broken contracts
            ----------------

            Forbidden contract one
            ----------------------

            mypackage.foo is not allowed to import mypackage.bar:

                mypackage.foo:8: from mypackage import bar
                mypackage.foo:16: from mypackage.bar import something
            """
        )

    def test_debug_mode_doesnt_swallow_exception(self):
        some_exception = RuntimeError("There was some sort of exception.")
        reader = ExceptionRaisingUserOptionReader(exception=some_exception)
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
            PRINTER=FakePrinter(),
        )

        with pytest.raises(some_exception.__class__, match=str(some_exception)):
            lint_imports(is_debug_mode=True)

    def test_non_debug_mode_prints_exception(self):
        some_exception = RuntimeError("There was some sort of exception.")
        reader = ExceptionRaisingUserOptionReader(exception=some_exception)
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
            PRINTER=FakePrinter(),
        )

        lint_imports(is_debug_mode=False)

        settings.PRINTER.pop_and_assert(
            """
            =============
            Import Linter
            =============

            There was some sort of exception.
            """
        )

    def _configure(
        self,
        contracts_options: List[Dict[str, Any]],
        session_options: Optional[Dict[str, Any]] = None,
        contract_types: Optional[List[str]] = None,
        graph: Optional[ImportGraph] = None,
        graph_builder: Optional[GraphBuilder] = None,
        timer: Optional[FakeTimer] = None,
    ):
        session_options = session_options or {"root_package": "mypackage"}
        if not contract_types:
            contract_types = [
                "always_passes: tests.helpers.contracts.AlwaysPassesContract",
                "always_fails: tests.helpers.contracts.AlwaysFailsContract",
                "fields: tests.helpers.contracts.FieldsContract",
                "forbidden: tests.helpers.contracts.ForbiddenImportContract",
                "noisy: tests.helpers.contracts.NoisyContract",
            ]
        session_options["contract_types"] = contract_types

        reader = FakeUserOptionReader(
            UserOptions(session_options=session_options, contracts_options=contracts_options)
        )
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=graph_builder or FakeGraphBuilder(),
            PRINTER=FakePrinter(),
            TIMER=timer or FakeTimer(),
            DEFAULT_CACHE_DIR=SOME_CACHE_DIR,
        )
        if graph is None:
            graph = self._build_default_graph()

        settings.GRAPH_BUILDER.inject_graph(graph)

    def _build_default_graph(self):
        graph = ImportGraph()

        # Add 26 modules.
        for letter in string.ascii_lowercase:
            graph.add_module(f"mypackage.{letter}")

        # Add 10 imports in total.
        for imported in ("d", "e", "f"):
            for importer in ("a", "b", "c"):
                graph.add_import(
                    importer=f"mypackage.{importer}", imported=f"mypackage.{imported}"
                )  # 3 * 3 = 9 imports.
        graph.add_import(importer="mypackage.d", imported="mypackage.f")  # 1 extra import.
        return graph


class TestMultipleRootPackages:
    def test_builder_is_called_with_root_packages(self):
        builder = FakeGraphBuilder()
        root_package_names = ["mypackageone", "mypackagetwo"]
        settings.configure(GRAPH_BUILDER=builder, PRINTER=FakePrinter())

        create_report(
            UserOptions(
                session_options={"root_packages": root_package_names},
                contracts_options=[],
            )
        )

        assert builder.build_arguments["root_package_names"] == root_package_names


class TestGraphCopying:
    def test_graph_can_be_mutated_without_affecting_other_contracts(self):
        # The MutationCheckContract checks that there are a certain number of modules and imports
        # in the graph, then adds one more module and one more import. We can check two such
        # contracts and the second one will fail, if the graph gets mutated by other contracts.
        session_options = {
            "root_package": "mypackage",
            "contract_types": ["mutation_check: tests.helpers.contracts.MutationCheckContract"],
        }

        reader = FakeUserOptionReader(
            UserOptions(
                session_options=session_options,
                contracts_options=[
                    {
                        "type": "mutation_check",
                        "name": "Contract one",
                        "number_of_modules": "5",
                        "number_of_imports": "2",
                    },
                    {
                        "type": "mutation_check",
                        "name": "Contract two",
                        "number_of_modules": "5",
                        "number_of_imports": "2",
                    },
                ],
            )
        )
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
            PRINTER=FakePrinter(),
        )

        graph = ImportGraph()

        # Create a graph with five modules and two imports.
        for module in ("one", "two", "three", "four", "five"):
            graph.add_module(module)
        graph.add_import(importer="one", imported="two")
        graph.add_import(importer="one", imported="three")

        settings.GRAPH_BUILDER.inject_graph(graph)

        result = lint_imports(is_debug_mode=True)

        assert result == SUCCESS


class TestCreateReport:
    @pytest.mark.parametrize(
        "limit_to_contracts, expected_message",
        (
            (
                ("a-contract-that-exists", "a-contract-that-doesnt-exist"),
                "Could not find contract 'a-contract-that-doesnt-exist'.\n\n"
                "You asked to limit the check to that contract, but nothing exists with that id.",
            ),
            (
                (
                    "a-contract-that-exists",
                    "a-contract-that-doesnt-exist",
                    "another-contract-that-doesnt-exist",
                ),
                "Could not find the following contract ids: "
                "a-contract-that-doesnt-exist, another-contract-that-doesnt-exist.\n\n"
                "You asked to limit the check to those contracts, but there are no "
                "contracts with those ids.",
            ),
        ),
    )
    def test_raises_exception_if_limited_to_nonexistent_contract(
        self, limit_to_contracts, expected_message
    ):
        settings.configure(GRAPH_BUILDER=FakeGraphBuilder(), PRINTER=FakePrinter())

        with pytest.raises(
            ValueError,
            match=re.escape(expected_message),
        ):
            create_report(
                user_options=UserOptions(
                    session_options={"root_packages": ["mypackage"]},
                    contracts_options=[
                        {
                            "type": "always_passes",
                            "id": "a-contract-that-exists",
                            "name": "Contract one",
                        },
                    ],
                ),
                limit_to_contracts=limit_to_contracts,
            )


class TestReadUserOptions:
    @pytest.mark.parametrize("filename", [".importlinter", "setup.cfg", "foo", "foo.bar"])
    def test_default_behavior(self, filename):
        expected_error = RuntimeError("expected")

        foo_reader = ExceptionRaisingUserOptionReader(AssertionError)
        ini_reader = ExceptionRaisingUserOptionReader(expected_error)
        toml_reader = ExceptionRaisingUserOptionReader(AssertionError)

        settings.configure(
            USER_OPTION_READERS={
                "foo": foo_reader,
                "ini": ini_reader,
                "toml": toml_reader,
            },
        )
        with pytest.raises(RuntimeError, match="expected"):
            lint_imports(filename, is_debug_mode=True)

    @pytest.mark.parametrize("filename", ["pyproject.toml", "foo.toml"])
    def test_toml_file(self, filename):
        expected_error = RuntimeError("expected")

        foo_reader = ExceptionRaisingUserOptionReader(AssertionError)
        ini_reader = ExceptionRaisingUserOptionReader(AssertionError)
        toml_reader = ExceptionRaisingUserOptionReader(expected_error)

        settings.configure(
            USER_OPTION_READERS={
                "foo": foo_reader,
                "ini": ini_reader,
                "toml": toml_reader,
            },
        )
        with pytest.raises(RuntimeError, match="expected"):
            lint_imports(filename, is_debug_mode=True)



import textwrap
from typing import Optional

from importlinter.application.ports.printing import Printer


class FakePrinter(Printer):
    def __init__(self) -> None:
        self._buffer = ""

    def print(
        self,
        text: str = "",
        bold: bool = False,
        color: Optional[str] = None,
        newline: bool = True,
    ) -> None:
        """
        Prints a line.
        """
        self._buffer += text
        if newline:
            self._buffer += "\n"

    def pop_and_assert(self, expected_string):
        """
        Assert that the string is what is in the buffer, removing it from the buffer.

        To aid with readable test assertions, the expected string will have leading and trailing
        lines removed, and the whole thing will be dedented, before comparison.
        """

        modified_expected_string = textwrap.dedent(expected_string.strip("\n"))

        popped_string = self._buffer
        self._buffer = ""

        assert popped_string == modified_expected_string



import abc
from typing import Optional


class Printer(abc.ABC):
    @abc.abstractmethod
    def print(
        self,
        text: str = "",
        bold: bool = False,
        color: Optional[str] = None,
        newline: bool = True,
    ) -> None:
        """
        Prints a line.
        """
        raise NotImplementedError



from typing import Optional

import click

from importlinter.application.ports.printing import Printer


class ClickPrinter(Printer):
    """
    Console printer that uses Click's formatting helpers.
    """

    def print(
        self,
        text: str = "",
        bold: bool = False,
        color: Optional[str] = None,
        newline: bool = True,
    ) -> None:
        # click.secho(text, bold=bold, fg=color, nl=newline)
        # A tricky solution to gh-267
        # Remove when click reaches v9.0.0
        print(click.style(text, bold=bold, fg=color), end=("\n" if newline else ""))



from typing import Tuple, List, Dict

import pytest
from grimp.adaptors.graph import ImportGraph

from importlinter.configuration import settings
from importlinter.contracts.forbidden import ForbiddenContract
from importlinter.domain.contract import ContractCheck
from tests.adapters.printing import FakePrinter
from tests.adapters.timing import FakeTimer


@pytest.fixture(scope="module", autouse=True)
def configure():
    settings.configure(TIMER=FakeTimer())


class TestForbiddenContract:
    @pytest.mark.parametrize("as_packages", ("False", "True"))
    def test_is_kept_when_no_forbidden_modules_imported(
        self,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", "mypackage.yellow"),
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    @pytest.mark.parametrize(
        "as_packages, forbidden_modules, expected_invalid_chains",
        [
            (
                True,
                (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.one",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.one.alpha",
                                    "imported": "mypackage.green.beta",
                                    "line_numbers": (3,),
                                }
                            ],
                            [
                                {
                                    "importer": "mypackage.one.alpha.circle",
                                    "imported": "mypackage.green.beta.sphere",
                                    "line_numbers": (8,),
                                },
                            ],
                        ],
                    },
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                }
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.purple",
                        "downstream_module": "mypackage.two",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.two",
                                    "imported": "mypackage.utils",
                                    "line_numbers": (9,),
                                },
                                {
                                    "importer": "mypackage.utils",
                                    "imported": "mypackage.purple",
                                    "line_numbers": (1,),
                                },
                            ]
                        ],
                    },
                ],
            ),
            (
                False,
                (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                }
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.purple",
                        "downstream_module": "mypackage.two",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.two",
                                    "imported": "mypackage.utils",
                                    "line_numbers": (9,),
                                },
                                {
                                    "importer": "mypackage.utils",
                                    "imported": "mypackage.purple",
                                    "line_numbers": (1,),
                                },
                            ]
                        ],
                    },
                ],
            ),
        ],
    )
    def test_is_broken_when_forbidden_modules_imported(
        self,
        as_packages: str,
        forbidden_modules: Tuple[str],
        expected_invalid_chains: List[Dict],
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=forbidden_modules,
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert not contract_check.kept

        expected_metadata = {"invalid_chains": expected_invalid_chains}

        assert expected_metadata == contract_check.metadata

    @pytest.mark.parametrize(
        "as_packages",
        (False, True),
    )
    def test_is_broken_when_forbidden_external_modules_imported(
        self,
        as_packages: str,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("sqlalchemy", "requests"),
            include_external_packages=True,
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "sqlalchemy",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "sqlalchemy",
                                "line_numbers": (1,),
                            }
                        ]
                    ],
                }
            ]
        }

        assert expected_metadata == contract_check.metadata

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_is_invalid_when_forbidden_externals_but_graph_does_not_include_externals(
        self,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("sqlalchemy", "requests"),
            as_packages=as_packages,
        )

        with pytest.raises(
            ValueError,
            match=(
                "The top level configuration must have include_external_packages=True "
                "when there are external forbidden modules."
            ),
        ):
            contract.check(graph=graph, verbose=False)

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_ignore_imports_tolerates_duplicates(self, as_packages: bool):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", "mypackage.yellow"),
            ignore_imports=(
                "mypackage.three -> mypackage.green",
                "mypackage.utils -> mypackage.purple",
                "mypackage.three -> mypackage.green",
            ),
            include_external_packages=False,
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.kept

    def test_wildcards_in_source_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green"),
            source_modules=("mypackage.one.*",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_recursive_wildcards_in_source_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green"),
            source_modules=("mypackage.one.**",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha.circle",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_wildcards_in_forbidden_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green.*"),
            source_modules=("mypackage.one",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green.beta",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_recursive_wildcards_in_forbidden_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green.**"),
            source_modules=("mypackage.one",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green.beta",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
                {
                    "upstream_module": "mypackage.green.beta.sphere",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    @pytest.mark.parametrize(
        "as_packages, forbidden_modules, expected_invalid_chains",
        [
            (
                True,
                ("mypackage.green",),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.one",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.one.alpha.circle",
                                    "imported": "mypackage.green.beta.sphere",
                                    "line_numbers": (8,),
                                },
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                },
                            ]
                        ],
                    },
                ],
            ),
            (
                False,
                ("mypackage.green",),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                },
                            ]
                        ],
                    },
                ],
            ),
        ],
    )
    def test_ignore_imports_with_wildcards(
        self,
        as_packages: str,
        forbidden_modules: Tuple[str],
        expected_invalid_chains: List[Dict],
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=forbidden_modules,
            ignore_imports=("mypackage.*.alpha -> mypackage.*.beta",),
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": expected_invalid_chains,
        }

    @pytest.mark.parametrize("as_packages", (False, True))
    def test_ignore_imports_with_recursive_wildcards(
        self,
        as_packages: str,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green",),
            ignore_imports=(
                "mypackage.**.alpha -> mypackage.**.beta",
                "mypackage.**.circle -> mypackage.**.sphere",
            ),
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "mypackage.green",
                                "line_numbers": (4,),
                            }
                        ]
                    ],
                },
            ]
        }

    @pytest.mark.parametrize(
        "importer",
        ("mypackage.one", "mypackage.one.alpha"),
    )
    @pytest.mark.parametrize(
        "imported",
        ("mypackage.mauve", "mypackage.mauve.beta"),
    )
    @pytest.mark.parametrize(
        "allow_indirect_imports",
        (False, True),
    )
    def test_allow_indirect_imports(self, importer, imported, allow_indirect_imports):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.mauve",),
            allow_indirect_imports=allow_indirect_imports,
        )
        graph.add_module("mypackage.mauve")
        # Add a direct import.
        graph.add_import(
            importer=importer,
            imported=imported,
            line_number=10,
            line_contents="-",
        )
        # Add an indirect import.
        graph.add_import(
            importer="mypackage.one.delta",
            imported="mypackage.something",
            line_number=20,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.something",
            imported="mypackage.mauve.gamma",
            line_number=30,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        direct_chain = [
            {
                "importer": importer,
                "imported": imported,
                "line_numbers": (10,),
            },
        ]
        indirect_chain = [
            {
                "importer": "mypackage.one.delta",
                "imported": "mypackage.something",
                "line_numbers": (20,),
            },
            {
                "importer": "mypackage.something",
                "imported": "mypackage.mauve.gamma",
                "line_numbers": (30,),
            },
        ]
        if allow_indirect_imports:
            expected_chains = [direct_chain]
        else:
            expected_chains = [direct_chain, indirect_chain]
        assert contract_check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.mauve",
                    "downstream_module": "mypackage.one",
                    "chains": expected_chains,
                },
            ],
        }

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_ignore_imports_adds_warnings(self, as_packages: bool):
        graph = self._build_graph()
        contract = ForbiddenContract(
            name="Forbid contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
                "forbidden_modules": "mypackage.purple",
                "ignore_imports": [
                    "mypackage.one -> mypackage.two.nonexistent",
                    "mypackage.*.nonexistent -> mypackage.three",
                ],
                "unmatched_ignore_imports_alerting": "warn",
                "as_packages": as_packages,
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert set(contract_check.warnings) == {
            "No matches for ignored import mypackage.one -> mypackage.two.nonexistent.",
            "No matches for ignored import mypackage.*.nonexistent -> mypackage.three.",
        }

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    @pytest.mark.parametrize(
        "module, expected_error",
        (
            (
                "requests.something",
                (
                    "Invalid forbidden module requests.something: "
                    "subpackages of external packages are not valid."
                ),
            ),
            # N.B. google.protobuf is a namespace package, but unless it is specified
            # as a root package, it will just appear in the graph as an external package
            # named 'google', so won't be treated any differently to other packages.
            (
                "google.protobuf",
                (
                    "Invalid forbidden module google.protobuf: "
                    "subpackages of external packages are not valid."
                ),
            ),
        ),
    )
    def test_is_invalid_when_subpackages_of_external_packages_are_provided(
        self,
        module: str,
        expected_error: str,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", module),
            include_external_packages=True,
            as_packages=as_packages,
        )

        with pytest.raises(
            ValueError,
            match=expected_error,
        ):
            contract.check(graph=graph, verbose=False)

    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "one",
            "one.alpha",
            "one.alpha.circle",
            "two",
            "three",
            "blue",
            "green",
            "green.beta",
            "green.beta.sphere",
            "yellow",
            "purple",
            "utils",
        ):
            graph.add_module(f"mypackage.{module}")
        for external_module in ("sqlalchemy", "requests", "google"):
            graph.add_module(external_module, is_squashed=True)
        graph.add_import(
            importer="mypackage.one.alpha",
            imported="mypackage.green.beta",
            line_number=3,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.one.alpha.circle",
            imported="mypackage.green.beta.sphere",
            line_number=8,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="mypackage.green",
            line_number=4,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.two",
            imported="mypackage.utils",
            line_number=9,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.utils",
            imported="mypackage.purple",
            line_number=1,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="sqlalchemy",
            line_number=1,
            line_contents="foo",
        )
        return graph

    def _build_contract(
        self,
        forbidden_modules,
        ignore_imports=None,
        include_external_packages=False,
        allow_indirect_imports=None,
        source_modules=None,
        as_packages=True,
    ):
        session_options = {"root_packages": ["mypackage"]}
        if include_external_packages:
            session_options["include_external_packages"] = "True"

        contract_options = {
            "source_modules": source_modules
            or ("mypackage.one", "mypackage.two", "mypackage.three"),
            "forbidden_modules": forbidden_modules,
            "ignore_imports": ignore_imports or [],
        }

        if allow_indirect_imports is not None:
            contract_options["allow_indirect_imports"] = (
                "true" if allow_indirect_imports else "false"
            )
        contract_options["as_packages"] = "true" if as_packages else "false"

        return ForbiddenContract(
            name="Forbid contract",
            session_options=session_options,
            contract_options=contract_options,
        )


class TestForbiddenContractForNamespacePackages:
    @pytest.mark.parametrize(
        "forbidden, is_kept",
        [
            ("namespace.subnamespace.portiontwo.green", False),
            ("namespace.subnamespace.portiontwo.blue", True),
        ],
    )
    def test_allows_forbidding_of_inter_portion_imports(self, forbidden, is_kept):
        graph = ImportGraph()
        for module in (
            "portionone",
            "portionone.blue",
            "subnamespace.portiontwo",
            "subnamespace.portiontwo.green",
            "subnamespace.portiontwo.blue",
        ):
            graph.add_module(f"namespace.{module}")
        for external_module in ("sqlalchemy", "requests"):
            graph.add_module(external_module, is_squashed=True)
        # Add import from one portion to another.
        graph.add_import(
            importer="namespace.portionone.blue",
            imported="namespace.subnamespace.portiontwo.green",
            line_number=3,
            line_contents="-",
        )
        contract = self._build_contract(
            root_packages=["namespace.portionone", "namespace.subnamespace.portiontwo"],
            source_modules=("namespace.portionone.blue",),
            forbidden_modules=(forbidden,),
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept == is_kept

    def _build_contract(
        self,
        root_packages,
        source_modules,
        forbidden_modules,
        include_external_packages=False,
    ):
        session_options = {"root_packages": root_packages}
        if include_external_packages:
            session_options["include_external_packages"] = "True"

        return ForbiddenContract(
            name="Forbid contract",
            session_options=session_options,
            contract_options={
                "source_modules": source_modules,
                "forbidden_modules": forbidden_modules,
            },
        )


def test_render_broken_contract():
    settings.configure(PRINTER=FakePrinter())
    contract = ForbiddenContract(
        name="Forbid contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
            "forbidden_modules": (
                "mypackage.blue",
                "mypackage.green",
                "mypackage.yellow",
                "mypackage.purple",
            ),
        },
    )
    check = ContractCheck(
        kept=False,
        metadata={
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.purple",
                    "downstream_module": "mypackage.two",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.two",
                                "imported": "mypackage.utils",
                                "line_numbers": (9,),
                            },
                            {
                                "importer": "mypackage.utils",
                                "imported": "mypackage.purple",
                                "line_numbers": (1,),
                            },
                        ]
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "mypackage.green",
                                "line_numbers": (4,),
                            }
                        ]
                    ],
                },
            ]
        },
    )

    contract.render_broken_contract(check)

    settings.PRINTER.pop_and_assert(
        """
        mypackage.two is not allowed to import mypackage.purple:

        -   mypackage.two -> mypackage.utils (l.9)
            mypackage.utils -> mypackage.purple (l.1)


        mypackage.three is not allowed to import mypackage.green:

        -   mypackage.three -> mypackage.green (l.4)


        """
    )


class TestVerbosePrint:
    def test_verbose(self):
        timer = FakeTimer()
        timer.setup(tick_duration=10, increment=0)
        settings.configure(PRINTER=FakePrinter(), TIMER=timer)

        graph = ImportGraph()
        for module in (
            "one",
            "one.alpha",
            "two",
            "three",
            "blue",
            "green",
            "green.beta",
            "yellow",
            "purple",
            "utils",
        ):
            graph.add_module(f"mypackage.{module}")
        graph.add_import(
            importer="mypackage.one.alpha",
            imported="mypackage.green.beta",
            line_number=3,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="mypackage.green",
            line_number=4,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.two",
            imported="mypackage.utils",
            line_number=9,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.utils",
            imported="mypackage.purple",
            line_number=1,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="sqlalchemy",
            line_number=1,
            line_contents="foo",
        )
        contract = ForbiddenContract(
            name="Forbid contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
                "forbidden_modules": (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                "ignore_imports": [],
                "allow_indirect_imports": "false",
            },
        )

        contract.check(graph=graph, verbose=True)

        settings.PRINTER.pop_and_assert(
            """
            Searching for import chains from mypackage.one to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.one to mypackage.green...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.one to mypackage.purple...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.one to mypackage.yellow...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.green...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.three to mypackage.purple...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.yellow...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.green...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.purple...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.two to mypackage.yellow...
            Found 0 illegal chains in 10s.
            """
        )
